from typing import Annotated  # noqa: D100

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

V3_FACTORY_ADDRESS = HexAddress(HexStr("0x1a3c9B1d2F0529D97f2afC5136Cc23e58f1FD35B"))


class GetPoolRequest(BaseModel):
    token_a: Annotated[HexAddress, Name("tokenA")]
    token_b: Annotated[HexAddress, Name("tokenB")]


class CamelotV3Factory(ProtocolBase):
    pool_by_pair: Annotated[
        ContractFunc[GetPoolRequest, HexAddress], Name("poolByPair")
    ] = METHOD



from typing import Annotated

from eth_rpc.types import Name, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class AddLiquidityRequest(BaseModel):
    token_a: Annotated[HexAddress, Name("tokenA")]
    token_b: Annotated[HexAddress, Name("tokenB")]
    amount_a_desired: Annotated[primitives.uint256, Name("amountADesired")]
    amount_b_desired: Annotated[primitives.uint256, Name("amountBDesired")]
    amount_a_min: Annotated[primitives.uint256, Name("amountAMin")]
    amount_b_min: Annotated[primitives.uint256, Name("amountBMin")]
    to: HexAddress
    deadline: primitives.uint256


class AddLiquidityETHRequest(BaseModel):
    token: Annotated[HexAddress, Name("token")]
    amount_token_desired: Annotated[primitives.uint256, Name("amountTokenDesired")]
    amount_token_min: Annotated[primitives.uint256, Name("amountTokenMin")]
    amount_eth_min: Annotated[primitives.uint256, Name("amountETHMin")]
    to: HexAddress
    deadline: primitives.uint256


class AddLiquidityResponse(BaseModel):
    amount_a: Annotated[primitives.uint256, Name("amountA")]
    amount_b: Annotated[primitives.uint256, Name("amountB")]
    liquidity: primitives.uint256


class RemoveLiquidityRequest(BaseModel):
    token_a: Annotated[primitives.address, Name("tokenA")]
    token_b: Annotated[primitives.address, Name("tokenB")]
    liquidity: primitives.uint256
    amount_a_min: Annotated[primitives.uint256, Name("amountAMin")]
    amount_b_min: Annotated[primitives.uint256, Name("amountBMin")]
    to: HexAddress
    deadline: primitives.uint256


class RemoveLiquidityETHRequest(BaseModel):
    token: Annotated[primitives.address, Name("token")]
    liquidity: primitives.uint256
    amount_token_min: Annotated[primitives.uint256, Name("amountTokenMin")]
    amount_eth_min: Annotated[primitives.uint256, Name("amountETHMin")]
    to: HexAddress
    deadline: primitives.uint256


class RemoveLiquidityResponse(BaseModel):
    amount_a: Annotated[primitives.uint256, Name("amountA")]
    amount_b: Annotated[primitives.uint256, Name("amountB")]


class RemoveLiquidityETHResponse(BaseModel):
    amount_token: Annotated[primitives.uint256, Name("amountToken")]
    amount_eth: Annotated[primitives.uint256, Name("amountETH")]



from typing import Annotated  # noqa: D100

from eth_rpc.contract import ContractFunc
from eth_rpc.types import METHOD, Name, NoArgs
from eth_typeshed.erc20 import ERC20
from eth_typing import HexAddress


class StakingPool(ERC20):
    pool_rewards: Annotated[ContractFunc[NoArgs, HexAddress], Name("poolRewards")] = (
        METHOD
    )
    index_fund: Annotated[ContractFunc[NoArgs, HexAddress], Name("indexFund")] = METHOD
    staking_token: Annotated[ContractFunc[NoArgs, HexAddress], Name("stakingToken")] = (
        METHOD
    )



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from eth_typeshed.erc20 import OwnerRequest
from eth_typeshed.multicall import multicall
from eth_typing import HexAddress, HexStr

from .position import OwnerTokenRequest, Position

NONFUNGIBLE_POSITION_MANAGER_ADDRESS = HexAddress(
    HexStr("0xC36442b4a4522E871399CD717aBDD847Ab11FE88")
)


class NonfungiblePositionManager(ProtocolBase):
    balance_of: Annotated[
        ContractFunc[
            OwnerRequest,
            Annotated[primitives.uint256, Name("amount")],
        ],
        Name("balanceOf"),
    ] = METHOD
    token_of_owner_by_index: Annotated[
        ContractFunc[
            OwnerTokenRequest,
            primitives.uint256,
        ],
        Name("tokenOfOwnerByIndex"),
    ] = METHOD
    positions: ContractFunc[
        primitives.uint256,
        Position,
    ] = METHOD

    async def get_all_indices(self, owner: HexAddress) -> list[int]:
        balance = await self.balance_of(OwnerRequest(owner=owner)).get_async()
        calls = []
        for i in range(balance):
            calls.append(
                self.token_of_owner_by_index(
                    OwnerTokenRequest(
                        owner=owner,
                        index=primitives.uint256(i),
                    )
                )
            )
        return await multicall.execute(*calls)

    async def get_all_positions(self, owner: HexAddress) -> list[Position]:
        indices = await self.get_all_indices(owner)
        return await multicall.execute(
            *[self.positions(primitives.uint256(index)) for index in indices]
        )



from .erc20 import ERC20, ERC20BytesMetadata
from .events import TransferEvent, TransferEventType
from .types import (
    ApproveRequest,
    OwnerRequest,
    OwnerSpenderRequest,
    TransferFromRequest,
    TransferRequest,
)

__all__ = [
    "ERC20",
    "ERC20BytesMetadata",
    "TransferEvent",
    "TransferEventType",
    "OwnerRequest",
    "OwnerSpenderRequest",
    "ApproveRequest",
    "TransferRequest",
    "TransferFromRequest",
]



from typing import Annotated, NamedTuple  # noqa: D100

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

PEAPOD_INDEX_MANAGER = HexAddress(HexStr("0x0BB39BA2EE60F825348676F9A87B7CD1E3B4AE6B"))


class IndexAndStatus(NamedTuple):
    index_address: HexAddress
    verified: primitives.bool


class IndexAndStatuses(BaseModel):
    values: list[IndexAndStatus]


class IndexManager(ProtocolBase):
    all_indexes: Annotated[
        ContractFunc[NoArgs, IndexAndStatuses],
        Name("allIndexes"),
    ] = METHOD
    authorized: ContractFunc[tuple[HexAddress], bool] = METHOD
    index_length: Annotated[
        ContractFunc[NoArgs, primitives.uint256], Name("indexLength")
    ] = METHOD
    indexes: ContractFunc[primitives.uint256, IndexAndStatus] = METHOD
    owner: ContractFunc[NoArgs, HexAddress] = METHOD



import math

from .tick import get_tick_at_sqrt_price

Q96 = 2**96
TICK_BASE = 1.0001


def price_to_tick(p):
    return math.floor(math.log(p, TICK_BASE))


def price_to_sqrtp(p):
    return int(math.sqrt(p) * Q96)


def sqrtp_to_price(sqrt_price):
    return (sqrt_price / Q96) ** 2


def liquidity0(amount, pa, pb):
    if pa > pb:
        pa, pb = pb, pa
    return (amount * (pa * pb) / Q96) / (pb - pa)


def liquidity1(amount, pa, pb):
    if pa > pb:
        pa, pb = pb, pa
    return amount * Q96 / (pb - pa)


def liquidity_to_token_amounts(liquidity, sqrt_price_x96, tick_low, tick_high):
    sqrt_ratio_a = math.sqrt(TICK_BASE**tick_low)
    sqrt_ratio_b = math.sqrt(TICK_BASE**tick_high)
    current_tick = get_tick_at_sqrt_price(sqrt_price_x96)
    sqrt_price = sqrt_price_x96 / Q96
    amount0 = 0
    amount1 = 0

    if current_tick < tick_low:
        amount0 = math.floor(
            liquidity * ((sqrt_ratio_b - sqrt_ratio_a) / (sqrt_ratio_a * sqrt_ratio_b))
        )
    elif current_tick >= tick_high:
        amount1 = math.floor(liquidity * (sqrt_ratio_b - sqrt_ratio_a))
    elif tick_low <= current_tick < tick_high:
        amount0 = math.floor(
            liquidity * ((sqrt_ratio_b - sqrt_price) / (sqrt_price * sqrt_ratio_b))
        )
        amount1 = math.floor(liquidity * (sqrt_price - sqrt_ratio_a))

    return [amount0, amount1]


def tick_to_price(tick):
    return TICK_BASE**tick



import math

from ..constants import Q96


def tick_to_price(tick, token0_decimals=18, token1_decimals=18) -> float:
    return 1.0001**tick * (10 ** (token0_decimals - token1_decimals))


def get_tick_at_sqrt_price(sqrt_price_x96) -> int:
    return math.floor(math.log((sqrt_price_x96 / Q96) ** 2) / math.log(1.0001))



from eth_rpc.types import BLOCK_STRINGS
from eth_typeshed.multicall import multicall


async def try_execute_with_setters(
    calls_with_setters, block_number: int | BLOCK_STRINGS = "latest"
):
    if not block_number:
        block_number = "latest"
    calls = [call for call, _ in calls_with_setters]
    results = await multicall.try_execute(*calls, block_number=block_number)
    for result, setter in zip(results, calls_with_setters):
        _, set_result = setter
        if (
            result.success
            and result.result != "0x0000000000000000000000000000000000000000"
        ):
            set_result(result.result)
